; Boot sector to load & run "kernel.bin" from a DOS filesystem
;
; Copyright (C) 1999 Thomas Novelli
;	Retro is distributed under the terms of the GNU LGPL, version 2 or
;	greater. (For the text of the license, see the file COPYING.LIB)
;
; Platform: Intel x86
;	Based on BOOTP by John S. Fine - thanks!  Thanks also to Linus
;	Torvalds, M. J. Vines, and Francois-Rene Rideau, for the tips &
;	examples.
;
; To do:
;	8086 compatibility...? (no 32-bit regs)
;	(If anyone is actually interested in an 8086 port of Retro)
;	See John's BOOTR...
;
;	Test/debug.  Make sure these conditions are met:
;		Split transfers on track boundaries and 64k boundaries
;		Large FATs won't overrun memory or otherwise fuck-up
;
;	Store "KERNEL  BIN" and any configurable stuff at fixed offsets so
;	they can be modified (like rdev in linux)
;
;	FAT32 support? (in a different version of the loader for 386+ only)
;
;	Bum code :)   (leave a little more room for expansion & test code)
;
;
; How it works:
;
; BIOS: 
;  1. Loads Sector 1, Track 0, Head 0 to address 7C00
;  2. Checks the last two bytes of the bootsector, which must contain
;     the signature 0xAA55, or it says "Operating system not found".
;  3. Loads DL with:
;	0x00 if booted from drive A
;	0x80 if booted from drive C
;	(Note: What about bootable CD's? 0x81, 0x82, etc?)
;  4. Jumps to 0000:7C00
;
; Boot loader:
;  1. Loads KERNEL.BIN at 7E00 and runs it
;
; Memory map:
;	 0000 to  03FF	; Real mode Interrupt Vector Table
;	 0400 to  049F	; BIOS data area
;	 7000 to  7BFF	; Stack for boot loader
;	 7C00 to  7DFF	; Boot sector
;	 7E00 to 20000	; Kernel (that's ~100k, make more room if needed)
;	20000 to ?????	; scratch area for FAT (free once kernel is loaded)
;			; (could be several hundred K on a large hard disk)

BITS 16

;Note: we could detect the boot address if that helps. CALL $+2 and POP..
;All BIOSes probably load at 7C00, but another loader could do otherwise.

%define BOOTSEG		0x7C0	; BIOS puts the boot sector at address 7C00
%define STACKSEG	0x700	; This area should be free for the stack...

;; Some debugging macros
%macro hexp 1	;; hexp <number>	;7-8 bytes
	pusha
	mov	dx, %1
	call	hex_print
	popa
%endmacro

%define hang	jmp short $

;; --------------------------------------------------------

	struc	DISK_PARM	;BIOS hard disk parameters
DP_cyl		resw	1
DP_heads	resb	1
		resb	5	;Obsolete
DP_ctrl		resb	1
		resb	5	;Obsolete
DP_sect		resb	1
		resb	1
	endstruc

	struc	PT_ENT		;Partition table entry
BootFlag	resb	1
BeginHead       resb	1
BeginSector     resb	1
BeginCyl        resb	1
SystemID        resb	1
EndHead         resb	1
EndSector       resb	1
EndCyl          resb	1
RelSectorLow    resw	1
RelSectorHigh   resw	1
NumSectorsLow   resw	1
NumSectorsHigh  resw	1
	endstruc

	struc	BB		;Partition Boot block
		resb	0xD	;Things we ignore (first 13 bytes of disk)
.clu		resb	1	;Sectors per cluster
.res		resw	1	;Reserved sectors
.fats		resb	1	;Number of FATs
.root		resw	1	;Root directory entries
		resb	3	;Things we ignore
.fat		resw	1	;Sectors per fat
	endstruc

max_xfer	equ	0x78	;Maximum number of sectors to transfer at once
				;Must be divisible by 8
;;;;;;;;;;;;;;;;;;;;;;;;;;

;> ax=0001	si=0000		*s: 0000  (cs, ss, ds, es, fs, gs)
;> bx=7C00	di=0500
;> cx=0001	bp=0000
;> dx=0000	sp=FFFE

	ORG	0x7C00

bb:	jmp short start
	nop

	times	0x3B	db	0

;--------------------------------------------------------------------

start	xor	eax, eax	;Initialize segments
	mov     ds, ax		;ds, ss = 0
	mov     ss, ax
	mov	sp, bb		;stack grows down from 7C00
	mov	bp, sp		;bp points to boot block

%ifdef HDLOAD
	mov byte [bdrive], 0x80	;Load from hard disk even on a floppy boot
%else
	mov	[bdrive], dl	;BIOS passes this along; save it.
	test	dl, 0x80	;Hard disk or floppy?
	jz	.floppy
%endif

;>  eax = 0

;; Hard disk (16/32-bit FAT, w/ MBR)
	mov	bp, 0x800
	mov	si, 0x80		;Read master boot record
	mov	es, si
	xor	di, di
	inc	di			;eax=0, es=0x80, di=1
	call	read_sectors
	
	mov	si, 0x9BE-16		;si = address of partition table
.10	add	si, byte 16
;hexp [si+BootFlag]
;hexp si
	cmp	[si+BootFlag], al	;(al=0)  ;Is partition active?
	je	.10
	jg	.ptn_error
	mov	eax, [si+RelSectorLow]	;LBN of partition
	mov	dh, [si+SystemID]
	mov	[fattype], dh
	call	read_sectors		;read partition's boot sector

	jmp short .common

.ptn_error
	mov	al, 'P'			; 'P'artition problem
	jmp	error

;; Floppy disk (12-bit FAT)
;> eax = 0
.floppy
	mov byte [fattype], 1		;Assume FAT12


;; The rest is the same...
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;>  eax = LBN of partition (0 on a floppy)

.common
	xor	ebx, ebx
	xor	edi, edi

	mov	bx, [bp+BB.res]
	add	eax, ebx		;LBN of FAT #1
	push	eax
	movzx	bx, [bp+BB.fats]	;# of FATs
	imul	bx, [bp+BB.fat]		;times FAT size (in sectors)

	mov	dl, [bp+BB.clu]		;Save these..
	mov	[bb+BB.clu], dl
	mov	dx, [bp+BB.fat]
	mov	[bb+BB.fat], dx

	;; Read root directory
	add	eax, ebx		;LBN of rootdir
	mov	di, [bp+BB.root]	;# of rootdir entries
	push	di
	dec	di
	shr	di, 4			;# of sectors in rootdir
	inc	di
	mov	bp, 0x2000
	mov	es, bp
	call	read_sectors
	add	eax, edi		;LBN of cluster 2

	;; scan rootdir for "KERNEL.BIN"
	pop	bx			;entries
;mov bx,100
	xor	di, di
.20	mov	si, file_name
	xor	ecx, ecx
	mov	cl, 11			;8+3 char DOS filename
%if 0
	push	bx
	xor	bx,bx
.25	mov	ah, 0xE
	mov	al, [es:di]
	inc	di
	int	10h
	loop	.25
	pop	bx
%else
	a32 rep cmpsb
	je	found
%endif

	add	cl, 21			;next entry...
	add	edi, ecx
	dec	bx
	jnz	.20

	mov	al, 'K'			; 'K'ernel missing or bungled
	jmp	error

found	;; load FAT at es:0 (in place of rootdir)
	mov	bx, [es:edi+0xF]	;starting cluster of file
	mov	di, [bb+BB.fat]
	xchg	eax, ebp		;ebp = LBN of cluster 2
	pop	eax			;LBN of FAT #1
	push	bx
	call	read_sectors

;; load file at 0x7E00
	push	es
	pop	fs
	mov	cx, 0x7E0
	mov	es, cx

;; Register state at beginning of loop:
;;	   [sp] = next cluster (was just in bx)
;;	     si = ????
;;	   es:0 = destination
;;	     bp = LBN of cluster 2
;;	     ds = 0
.10
;	xor	eax, eax
;	xor	esi,esi
	pop	si			;get next cluster
;hexp si ;Cluster #
	mov	bx, si			;copy, leave si untouched
	shl	bx, 1			; *2  (for a 16-bit FAT entry)
	cmp byte [fattype], 1		;12-bit FAT?
	jne	.fat16
.fat12	cmp	si, 0xFF0		;Valid cluster #?
	jae	eof
	add	bx, si			; *3
	shr	bx, 1			; /2
	mov	bx, [fs:bx]		;get following cluster
;bum code: use RCR to set carry if odd, like in BOOTR
	test	si,1
	jz	.even
.odd	shr	bx, 4			;clear unused nibble for 12-bit
.even	and	bh, 0xF
;.even	and	bx, 0xFFF	;bummed
	jmp short .20
.fat16	cmp	si, 0xFFF0		;Valid cluster #?
	jae	eof
	mov	bx, [fs:bx]		;get following cluster
.20	push	bx
	dec	si
	dec	si
	movzx	ax, [bb+BB.clu]		;cluster size (in sectors)
	mov	di, ax
	mul	si
	add	ax, bp			;add LBN of cluster 2
	call	read_sectors

	mov	ax, es
	mov	si, di
	shl	si, 5			;cluster size (in 16-byte units)
	add	ax, si
	mov	es, ax
	jmp short .10


eof:
	call	motor_off
	jmp	kernel			;***************************

;--------------------------------------------------------------------

motor_off:
	mov	dx,03F2h		;Shut off floppy motor
	xor	al,al
	out	dx,al
	ret

read_sectors:	;; borrowed from John Fine's BOOTP example
;
; Input:
;	EAX = LBN
;	DI  = sector count
;	ES = segment
; Output:
;	BL = low byte of ES
;	EBX high half cleared
;	DL = [bdrive]	(drive A=0x00  C=0x80)
;	EDX high half cleared
;	ESI = 0
; Clobbered:
;	BH, CX, DH
;
	push	eax
	push	di
	push	es

.10:	push	eax		;LBN

	mov	cl, [bdrive]
	push	ds
	lds	si, [0x104]	;Hard drive 0 parameters
	cdq			;edx = 0
	movzx	ebx, byte [si+DP_sect]	;HD
	test	cl, 0x80	;floppy?
	jnz	.12
	mov	bl, 18		;then assume 18 sectors (1.44M disk)
.12	div	ebx		;EAX=track ;EDX=sector-1
	mov	bl, [si+DP_heads]	;HD
	test	cl, 0x80	;floppy?
	jnz	.13
	mov	bl, 2		;then assume 2 heads (hardcoded)
.13	pop	ds
	mov	cx, dx		;CL=sector-1 ;CH=0
	inc	cx		;CL=Sector number
	xor	dx, dx
	div	ebx

	mov	dh, dl		;Head
	mov	dl, [bdrive]	;Drive
	xchg	ch, al		;CH=Low 8 bits of cylinder number; AL=0
	shr	ax, 2		;AL[6:7]=High two bits of cylinder; AH=0
	or	cl, al		;CX = Cylinder and sector
	mov	al, max_xfer	;AX = Maximum sectors to xfer
	cmp	ax, di		;Total is larger?
	jb	.20		;Yes: transfer maximum
	xchg	ax, di		;No:  AX=total ;DI=maximum
.20:	mov	ah, 2		;Read
	xor	bx, bx
	int	13h
	jc near	disk_error

	mov	bx, es
	add	bh, max_xfer>>3	;Advance segment
	mov	es, bx

	pop	eax
	add	eax, byte max_xfer
	sub	di, byte max_xfer
	ja	.10

	pop	es
	pop	di
	pop	eax
	xor	esi, esi
	ret	

disk_error
	mov	al, 'D'		; 'D'isk error

error	mov	ah, 0xE
	xor	bx, bx
	int	0x10
	call	motor_off
	hang

;--------------------------------------------------------------------

%if 1
hex_print:
;
; Prints the number in dx
;
;				;16bit	32bit
        mov     cx,4		;cx,4	cx,8
print_digit:
        rol     dx,4		;dx,4	edx,8
        mov     ax,0E0Fh
        and     al,dl
        add     al,90h
        daa
        adc     al,40h
        daa
        int     10h
        loop    print_digit
	mov	al,32		;Print a space
	int	10h
        ret
%endif

;--------------------------------------------------------------------

;boot_msg	db "RETRO #9 bootloader (dosboot)",13,10,0

;disk_error_msg	db 13,10,"Disk Error",13,10,0



;Disk specs
;sectors:        dw      0
;tracks:		dw	80
;heads:		dw	2

;DiskSizes	db 36,18,15,9

;Variables - these locations should be unused..
bdrive	equ	0x0FFE
fattype	equ	0x0FFF

	times	512-11-2-($-$$)	db 0	; skip to end of sector
file_name db 'KERNEL  BIN'
	dw	0xAA55		; boot sector signature

kernel:	;; begins here, at 7E00
