; graphics.asm - Graphics primitives

GRAPHMODE       ; ( addr -- , Use font table at addr )
        push    eax

        upop    eax
        push    ebp
        mov     bp, ax

        xor     ah, ah
        mov     al, 0x13
        int16   0x10

        mov     ax, ds
        mov     es, ax
        mov     ax, 0x1120
        int16   0x10                    ;Set 8x8 graphics font table

        mov     ax, [vid.addr]
        mov     es, ax

        pop     ebp
        pop     eax
        ret

;Video variables
vid
.addr   dc      0xA000          ;Address
.w      dc      320             ;Width
.h      dc      200             ;Height
end GRAPHMODE



TEXTMODE
        xor     ah, ah
        mov     al, 0x03
        int16   0x10
        ret
end TEXTMODE



HLINE           ; ( c x1 x2 y -- , Draw a horizontal line )
        push    eax
        push    ebx
        push    ecx
        push    edi

        mov     ax, [vid.addr]
        mov     es, ax

        mov     eax, [vid.w]
        upsh    eax
        call    MULTIPLY
        upop    ebx             ; Row
        call    OVER
        call    MINUS
        upop    ecx             ; Start/end column
        upop    edi
        add     edi, ebx
        upop    eax             ; Color
        rep stosb

        pop     edi
        pop     ecx
        pop     ebx
        pop     eax
        ret
end HLINE



VLINE           ; ( c x y1 y2 -- , Draw a vertical line )
        ret
end VLINE



LINE            ; ( c x1 y1 x2 y2 -- , Draw a diagonal line )
;
;Algorithm:
; Use fixed-point numbers (16.16 in 32-bit, 10.6 in 16-bit)
; Start at topmost point
; Calculate dx (dy=1)
; Count = |y2-y1| (vertical distance)
; (x,y) = initial point (in fixed-point notation)
; Loop:
;   hline from x to x+dx (convert to integer)
;   x = x+dx
;   y++
;
        push    eax
        push    ebx
        push    ecx
        push    edx
        push    esi
        push    edi

        mov     ax, [vid.addr]
        mov     es, ax

        call    ROT             ;( c x1 x2 y2 y1 )
        upop    eax             ;y1
        upop    ecx             ;y2

        ; Find topmost point    AX = initial y-value
        cmp     eax, ecx
        jb      .noswap

        xchg    eax, ecx
        call    SWAP            ;( c x2 x1 )
.noswap

        inc     ecx
        sub     ecx, eax        ;CX = count = vertical distance

        mul     cell [vid.w]
        mov     esi, eax        ;SI = initial scanline (y * screen-width)

        upop    eax             ;Final x-value
        inc     eax
        upop    edi             ;Initial x-value

        sub     eax, edi        ;Calculate dx (in fixed-point)
        xor     edx, edx
        shl     eax, FIXSHIFT
        idiv    ecx
        mov     ebx, eax        ;BX = dx (delta-x)

        mov     edx, edi
        shl     edx, FIXSHIFT   ;DX = current x-value (fixed-point)

        upop    eax             ;AL = color

        cmp     edx, 0
        jge     .positive_dx
        std     ;If negative, EDI must be decremented during REP STOSB
.positive_dx

        add     edi, esi        ;DI = current point address
.loop
        push    ecx             ;Draw hline
        ; This next piece may seem unnecessary; maybe it is. It's for accuracy.
        mov     ecx, edx        ; Get x
        add     edx, ebx        ; Add dx
        sub     edx, ecx        ; Find difference
        xchg    edx, ecx
        shr     ecx, FIXSHIFT
        rep     stosb
        pop     ecx

        add     edi, [vid.w]
        loop    .loop

        pop     edi
        pop     esi
        pop     edx
        pop     ecx
        pop     ebx
        pop     eax
        ret
end LINE



PUTPIXEL        ; ( c x y -- , Draw a pixel )
        push    eax
        push    edi

        mov     ax, [vid.addr]
        mov     es, ax

        mov     eax, [vid.w]
        upsh    eax
        call    MULTIPLY
        call    PLUS
        upop    edi
        upop    eax
        mov     [es:edi], al

        pop     edi
        pop     eax
        ret
end PUTPIXEL



CIRCLE          ; ( c x y r -- , Draw a circle, radius=r at x,y )
;
; Variables:
;  cx  = DI
;  cy  = SI
;  df  = DX
;  de  = AX
;  dse = CX
;  temp = BX (for pixel plotting)
;  x = S1
;  y = S0
;  c = S2
;
; Optimize: Precalculate address (for cy, at least?)
;
        push    eax
        push    ebx
        push    ecx
        push    edx
        push    esi
        push    edi

        mov     ax, [vid.addr]
        mov     es, ax

        ;Initialize variables
        sub     di, di          ;cx = 0
        upop    si              ;cy = radius
        mov     ax, 3           ;de = 0

        mov     dx, si
        neg     dx
        inc     dx              ;df = 1-r = -r+1

        mov     cx, si
        sal     cx, 1           ;dse = 2r
        neg     cx
        add     cx, 5           ;dse = -2r+5

.loop
        ;Plot 8 points (1 per octant of circle)
        push    eax
        push    ecx

;Note: This is currently a text-mode circle plot :-)
;
        mov     ebx, S1         ;X
        add     ebx, edi
        mov     ecx, S0         ;Y
        add     ecx, esi
        upsh    ebx
        upsh    ecx
        call    AT_XY
        upsh    cell '*'
        call    EMIT

        mov     ebx, S1         ;X
        sub     ebx, edi
        mov     ecx, S0         ;Y
        add     ecx, esi
        upsh    ebx
        upsh    ecx
        call    AT_XY
        upsh    cell '*'
        call    EMIT

        pop     ecx
        pop     eax


        ;Update variables
        cmp     dx, 0           ;if (df < 0)
        jge     .greater
.less
        add     dx, ax          ;df+=de
        add     cx, 2           ;dse+=2
        jmp     .endif
.greater
        add     dx, cx          ;df+=dse
        add     cx, 4           ;dse+=4
        dec     si              ;cy--
.endif
        add     ax, 2           ;de+=2
        inc     di              ;cx++

        cmp     di, si          ;Repeat loop while (cx <= cy)
        jle near .loop

        pop     edi
        pop     esi
        pop     edx
        pop     ecx
        pop     ebx
        pop     eax
        ret
end CIRCLE
